@@ -36,12 +36,13 @@ namespace provisioner {
 
     void Services::registerHandlers(drogon::HttpAppFramework &app)
     {
-        app.registerHandler("/services", [](const HttpRequestPtr &req, std::function<void(const HttpResponsePtr &)> &&callback) {
+        // Add async services endpoint that does the heavy lifting
+        app.registerHandler("/api/v2/services", [](const HttpRequestPtr &req, std::function<void(const HttpResponsePtr &)> &&callback) {
             using namespace trantor;
-            LOG_INFO << "Services::services";
+            LOG_INFO << "Services::api-services (async)";
             
             // Add audit log entry for handler access
-            AuditLog::logHandlerAccess(req, "/services");
+            AuditLog::logHandlerAccess(req, "/api/v2/services");
 
             std::vector<ServiceInfo> serviceInfos;
             sd_bus *bus = nullptr;
@@ -311,83 +312,60 @@ namespace provisioner {
             
             sd_bus_unref(bus);
             
-            auto resp = drogon::HttpResponse::newHttpResponse();
+            // Always return JSON response for this API endpoint
+            Json::Value serviceArray(Json::arrayValue);
             
-            // Check the Accept header for JSON
-            auto acceptHeader = req->getHeader("Accept");
-            if (!acceptHeader.empty() && acceptHeader.find("application/json") != std::string::npos) {
-                // Return JSON response
-                Json::Value serviceArray(Json::arrayValue);
+            for (const auto& info : serviceInfos) {
+                Json::Value serviceObj;
+                serviceObj["name"] = info.name;
+                serviceObj["status"] = info.status;
+                serviceObj["active"] = info.active;
+                serviceObj["instance"] = info.instance;
+                serviceObj["base_name"] = info.base_name;
                 
-                for (const auto& info : serviceInfos) {
-                    Json::Value serviceObj;
-                    serviceObj["name"] = info.name;
-                    serviceObj["status"] = info.status;
-                    serviceObj["active"] = info.active;
-                    serviceObj["instance"] = info.instance;
-                    serviceObj["base_name"] = info.base_name;
-                    
-                    // Reconstruct the full service name for log links
-                    std::string fullName;
-                    if (!info.instance.empty()) {
-                        fullName = info.name + info.instance + ".service";
-                    } else {
-                        fullName = info.name + ".service";
-                    }
-                    serviceObj["full_name"] = fullName;
-                    
-                    // Removed per-service JSON logging
-                    
-                    serviceArray.append(serviceObj);
+                // Add timestamp in microseconds since epoch (will be converted to local time in frontend)
+                serviceObj["timestamp"] = Json::Value::UInt64(info.timestamp);
+                
+                // Reconstruct the full service name for log links
+                std::string fullName;
+                if (!info.instance.empty()) {
+                    fullName = info.name + info.instance + ".service";
+                } else {
+                    fullName = info.name + ".service";
                 }
+                serviceObj["full_name"] = fullName;
                 
-                // Minimal logging for JSON response
-                
-                Json::Value rootObj;
-                rootObj["services"] = serviceArray;
-                
-                resp->setStatusCode(drogon::k200OK);
-                resp->setContentTypeCode(drogon::CT_APPLICATION_JSON);
-                std::string jsonOutput = rootObj.toStyledString();
-                // Removed JSON output logging
-                resp->setBody(jsonOutput);
-                
-                LOG_INFO << "JSON response prepared with " << serviceArray.size() << " services";
-            } else {
-                // Return HTML view
-                drogon::HttpViewData viewData;
-                std::vector<std::map<std::string, std::string>> serviceMaps;
-                
-                for (auto && info : serviceInfos) {
-                    std::map<std::string, std::string> serviceMap;
-                    
-                    serviceMap["name"] = info.name;
-                    serviceMap["status"] = info.status;
-                    serviceMap["active"] = info.active;
-                    serviceMap["instance"] = info.instance;
-                    serviceMap["base_name"] = info.base_name;
-                    
-                    // Reconstruct the full service name for log links
-                    std::string fullName;
-                    if (!info.instance.empty()) {
-                        fullName = info.name + info.instance + ".service";
-                    } else {
-                        fullName = info.name + ".service";
-                    }
-                    serviceMap["full_name"] = fullName;
-                    
-                    // Removed per-service UI logging
-                    
-                    serviceMaps.push_back(serviceMap);
-                }
-                viewData.insert("services", serviceMaps);
-                viewData.insert("currentPage", std::string("services"));
-                // Minimal view data logging
-                resp = drogon::HttpResponse::newHttpViewResponse("services.csp", viewData);
-                
-                LOG_INFO << "HTML view prepared with " << serviceMaps.size() << " services";
+                serviceArray.append(serviceObj);
             }
             
+            Json::Value rootObj;
+            rootObj["services"] = serviceArray;
+            
+            auto resp = drogon::HttpResponse::newHttpJsonResponse(rootObj);
+            LOG_INFO << "Async JSON response prepared with " << serviceArray.size() << " services";
+            
+            callback(resp);
+        });
+
+        // Add fast HTML-only services page that shows loading state
+        app.registerHandler("/services", [](const HttpRequestPtr &req, std::function<void(const HttpResponsePtr &)> &&callback) {
+            using namespace trantor;
+            LOG_INFO << "Services::services (fast HTML)";
+            
+            // Add audit log entry for handler access
+            AuditLog::logHandlerAccess(req, "/services");
+
+            // Return HTML view immediately with empty services list
+            drogon::HttpViewData viewData;
+            std::vector<std::map<std::string, std::string>> serviceMaps; // Empty list
+            
+            viewData.insert("services", serviceMaps);
+            viewData.insert("currentPage", std::string("services"));
+            viewData.insert("loading", true); // Flag to show loading state
+            
+            auto resp = drogon::HttpResponse::newHttpViewResponse("services.csp", viewData);
+            LOG_INFO << "Fast HTML view prepared with loading state";
+            
             callback(resp);
         });
 
@@ -419,6 +397,39 @@ namespace provisioner {
             // Log systemd log access to audit log
             AuditLog::logSystemdAccess(serviceName);
 
+            // Parse pagination and ordering parameters
+            int page = 1;
+            int pageSize = 50;
+            std::string order = "desc"; // Default to newest first
+            
+            auto pageParam = req->getParameter("page");
+            if (!pageParam.empty()) {
+                try {
+                    page = std::stoi(pageParam);
+                    if (page < 1) page = 1;
+                } catch (...) {
+                    page = 1;
+                }
+            }
+            
+            auto pageSizeParam = req->getParameter("page_size");
+            if (!pageSizeParam.empty()) {
+                try {
+                    pageSize = std::stoi(pageSizeParam);
+                    if (pageSize < 1) pageSize = 50;
+                    if (pageSize > 500) pageSize = 500; // Cap at 500
+                } catch (...) {
+                    pageSize = 50;
+                }
+            }
+            
+            auto orderParam = req->getParameter("order");
+            if (!orderParam.empty() && (orderParam == "asc" || orderParam == "desc")) {
+                order = orderParam;
+            }
+            
+            LOG_INFO << "Fetching logs: page=" << page << ", page_size=" << pageSize << ", order=" << order;
+
             // Open the journal
             sd_journal *j;
             int r = sd_journal_open(&j, SD_JOURNAL_LOCAL_ONLY);
@@ -454,43 +465,43 @@ namespace provisioner {
                 return;
             }
             
-            // Seek to the end
-            r = sd_journal_seek_tail(j);
-            if (r < 0) {
-                LOG_ERROR << "Failed to seek to end of journal: " << strerror(-r);
-                sd_journal_close(j);
-                auto resp = provisioner::utils::createErrorResponse(
-                    req,
-                    "Failed to seek in journal",
-                    drogon::k500InternalServerError, 
-                    "Journal Error",
-                    "JOURNAL_SEEK_ERROR",
-                    std::string("Error: ") + strerror(-r)
-                );
-                callback(resp);
-                return;
+            // Count total entries for pagination metadata
+            int totalEntries = 0;
+            if (order == "desc") {
+                sd_journal_seek_head(j);
+            } else {
+                sd_journal_seek_tail(j);
             }
             
-            // Move back 200 log entries
-            r = sd_journal_previous_skip(j, 200);
-            if (r < 0) {
-                LOG_ERROR << "Failed to skip backwards in journal: " << strerror(-r);
-                sd_journal_close(j);
-                auto resp = provisioner::utils::createErrorResponse(
-                    req,
-                    "Failed to navigate journal",
-                    drogon::k500InternalServerError,
-                    "Journal Error", 
-                    "JOURNAL_NAVIGATION_ERROR",
-                    std::string("Error: ") + strerror(-r)
-                );
-                callback(resp);
-                return;
+            while ((order == "desc" ? sd_journal_next(j) : sd_journal_previous(j)) > 0) {
+                totalEntries++;
             }
             
-            // Collect log entries
+            int totalPages = (totalEntries + pageSize - 1) / pageSize;
+            if (page > totalPages && totalPages > 0) page = totalPages;
+            
+            // Seek to appropriate position based on order and page
+            if (order == "desc") {
+                // Newest first: start from tail, go backwards
+                sd_journal_seek_tail(j);
+                int skipCount = (page - 1) * pageSize;
+                if (skipCount > 0) {
+                    sd_journal_previous_skip(j, skipCount);
+                }
+            } else {
+                // Oldest first: start from head, go forwards
+                sd_journal_seek_head(j);
+                int skipCount = (page - 1) * pageSize;
+                if (skipCount > 0) {
+                    sd_journal_next_skip(j, skipCount);
+                }
+            }
+            
+            // Collect log entries for the current page
             std::vector<std::string> logEntries;
-            while (sd_journal_next(j) > 0) {
+            int collected = 0;
+            
+            while (collected < pageSize && (order == "desc" ? sd_journal_previous(j) : sd_journal_next(j)) > 0) {
                 const void *data;
                 size_t length;
                 
@@ -515,20 +526,28 @@ namespace provisioner {
                     continue;
                 }
                 
-                // Convert timestamp to readable format
+                // Convert timestamp to ISO 8601 format (UTC)
                 time_t secs = time / 1000000;
+                uint64_t usecs = time % 1000000;
                 struct tm tm;
-                localtime_r(&secs, &tm);
+                gmtime_r(&secs, &tm);
                 char timestr[64];
-                strftime(timestr, sizeof(timestr), "%Y-%m-%d %H:%M:%S", &tm);
+                strftime(timestr, sizeof(timestr), "%Y-%m-%dT%H:%M:%S", &tm);
+                
+                // Add milliseconds and Z suffix for ISO 8601 UTC format
+                char isotimestr[80];
+                snprintf(isotimestr, sizeof(isotimestr), "%s.%03ldZ", timestr, usecs / 1000);
                 
                 // Add the formatted log entry
-                std::string entry = std::string(timestr) + " " + message;
+                std::string entry = std::string(isotimestr) + " " + message;
                 logEntries.push_back(entry);
+                collected++;
             }
             
-            // Reverse the log entries to get reverse-chronological order (newest first)
-            std::reverse(logEntries.begin(), logEntries.end());
+            // If order is desc, we collected backwards, so reverse to show newest first
+            if (order == "desc") {
+                std::reverse(logEntries.begin(), logEntries.end());
+            }
             
             sd_journal_close(j);
             
@@ -548,6 +567,11 @@ namespace provisioner {
                 
                 root["logs"] = logArray;
                 root["service_name"] = serviceName;
+                root["page"] = page;
+                root["page_size"] = pageSize;
+                root["total_entries"] = totalEntries;
+                root["total_pages"] = totalPages;
+                root["order"] = order;
                 
                 resp->setStatusCode(drogon::k200OK);
                 resp->setContentTypeCode(drogon::CT_APPLICATION_JSON);
@@ -559,6 +583,11 @@ namespace provisioner {
                 viewData.insert("log_entries", logEntries);
                 viewData.insert("currentPage", std::string("services"));
                 viewData.insert("auto_refresh", true); // Flag to enable auto-refresh in template
+                viewData.insert("page", page);
+                viewData.insert("page_size", pageSize);
+                viewData.insert("total_entries", totalEntries);
+                viewData.insert("total_pages", totalPages);
+                viewData.insert("order", order);
                 resp = drogon::HttpResponse::newHttpViewResponse("service_log.csp", viewData);
             }
             
@@ -595,6 +624,39 @@ namespace provisioner {
             // Log systemd log access to audit log
             AuditLog::logSystemdAccess(serviceName);
 
+            // Parse pagination and ordering parameters
+            int page = 1;
+            int pageSize = 50;
+            std::string order = "desc"; // Default to newest first
+            
+            auto pageParam = req->getParameter("page");
+            if (!pageParam.empty()) {
+                try {
+                    page = std::stoi(pageParam);
+                    if (page < 1) page = 1;
+                } catch (...) {
+                    page = 1;
+                }
+            }
+            
+            auto pageSizeParam = req->getParameter("page_size");
+            if (!pageSizeParam.empty()) {
+                try {
+                    pageSize = std::stoi(pageSizeParam);
+                    if (pageSize < 1) pageSize = 50;
+                    if (pageSize > 500) pageSize = 500; // Cap at 500
+                } catch (...) {
+                    pageSize = 50;
+                }
+            }
+            
+            auto orderParam = req->getParameter("order");
+            if (!orderParam.empty() && (orderParam == "asc" || orderParam == "desc")) {
+                order = orderParam;
+            }
+            
+            LOG_INFO << "API fetching logs: page=" << page << ", page_size=" << pageSize << ", order=" << order;
+
             // Open the journal
             sd_journal *j;
             int r = sd_journal_open(&j, SD_JOURNAL_LOCAL_ONLY);
@@ -630,43 +692,43 @@ namespace provisioner {
                 return;
             }
             
-            // Seek to the end
-            r = sd_journal_seek_tail(j);
-            if (r < 0) {
-                LOG_ERROR << "Failed to seek to end of journal: " << strerror(-r);
-                sd_journal_close(j);
-                auto resp = provisioner::utils::createErrorResponse(
-                    req,
-                    "Failed to seek in journal",
-                    drogon::k500InternalServerError,
-                    "Journal Error",
-                    "JOURNAL_SEEK_ERROR",
-                    std::string("Error: ") + strerror(-r)
-                );
-                callback(resp);
-                return;
+            // Count total entries for pagination metadata
+            int totalEntries = 0;
+            if (order == "desc") {
+                sd_journal_seek_head(j);
+            } else {
+                sd_journal_seek_tail(j);
             }
             
-            // Move back 100 log entries for the API (lighter than the full view)
-            r = sd_journal_previous_skip(j, 100);
-            if (r < 0) {
-                LOG_ERROR << "Failed to skip backwards in journal: " << strerror(-r);
-                sd_journal_close(j);
-                auto resp = provisioner::utils::createErrorResponse(
-                    req,
-                    "Failed to navigate journal",
-                    drogon::k500InternalServerError,
-                    "Journal Error",
-                    "JOURNAL_NAVIGATION_ERROR",
-                    std::string("Error: ") + strerror(-r)
-                );
-                callback(resp);
-                return;
+            while ((order == "desc" ? sd_journal_next(j) : sd_journal_previous(j)) > 0) {
+                totalEntries++;
             }
             
-            // Collect log entries
+            int totalPages = (totalEntries + pageSize - 1) / pageSize;
+            if (page > totalPages && totalPages > 0) page = totalPages;
+            
+            // Seek to appropriate position based on order and page
+            if (order == "desc") {
+                // Newest first: start from tail, go backwards
+                sd_journal_seek_tail(j);
+                int skipCount = (page - 1) * pageSize;
+                if (skipCount > 0) {
+                    sd_journal_previous_skip(j, skipCount);
+                }
+            } else {
+                // Oldest first: start from head, go forwards
+                sd_journal_seek_head(j);
+                int skipCount = (page - 1) * pageSize;
+                if (skipCount > 0) {
+                    sd_journal_next_skip(j, skipCount);
+                }
+            }
+            
+            // Collect log entries for the current page
             std::vector<std::string> logEntries;
-            while (sd_journal_next(j) > 0) {
+            int collected = 0;
+            
+            while (collected < pageSize && (order == "desc" ? sd_journal_previous(j) : sd_journal_next(j)) > 0) {
                 const void *data;
                 size_t length;
                 
@@ -689,20 +751,28 @@ namespace provisioner {
                     continue;
                 }
                 
-                // Convert timestamp to readable format
+                // Convert timestamp to ISO 8601 format (UTC)
                 time_t secs = time / 1000000;
+                uint64_t usecs = time % 1000000;
                 struct tm tm;
-                localtime_r(&secs, &tm);
+                gmtime_r(&secs, &tm);
                 char timestr[64];
-                strftime(timestr, sizeof(timestr), "%Y-%m-%d %H:%M:%S", &tm);
+                strftime(timestr, sizeof(timestr), "%Y-%m-%dT%H:%M:%S", &tm);
+                
+                // Add milliseconds and Z suffix for ISO 8601 UTC format
+                char isotimestr[80];
+                snprintf(isotimestr, sizeof(isotimestr), "%s.%03ldZ", timestr, usecs / 1000);
                 
                 // Add the formatted log entry
-                std::string entry = std::string(timestr) + " " + message;
+                std::string entry = std::string(isotimestr) + " " + message;
                 logEntries.push_back(entry);
+                collected++;
             }
             
-            // Reverse the entries to get newest first
-            std::reverse(logEntries.begin(), logEntries.end());
+            // If order is desc, we collected backwards, so reverse to show newest first
+            if (order == "desc") {
+                std::reverse(logEntries.begin(), logEntries.end());
+            }
             
             sd_journal_close(j);
             
@@ -716,6 +786,11 @@ namespace provisioner {
             
             root["logs"] = logArray;
             root["service_name"] = serviceName;
+            root["page"] = page;
+            root["page_size"] = pageSize;
+            root["total_entries"] = totalEntries;
+            root["total_pages"] = totalPages;
+            root["order"] = order;
             
             auto resp = drogon::HttpResponse::newHttpJsonResponse(root);
             callback(resp);
