@@ -91,7 +91,8 @@ TARGET_DEVICE_PATH="$1"
 # - Y is the port number on that bus  
 # - Z is the optional port number for devices behind a hub
 # Example: "1-1.2" means bus 1, hub port 1, hub downstream port 2
-TARGET_USB_PATH="$(udevadm info "${TARGET_DEVICE_PATH}" | grep -oP '^M: \K.*')"
+# Extract USB device path from udevadm (P: line) and parse out the topology path
+TARGET_USB_PATH="$(udevadm info "${TARGET_DEVICE_PATH}" | grep -oP '^P: \K.*' | grep -oE '[0-9]+-[0-9]+(\.[0-9]+)*$' || true)"
 TARGET_DEVICE_FAMILY="$(udevadm info --name="$TARGET_DEVICE_PATH" --query=property --property=ID_MODEL_ID --value)"
 # TARGET_DEVICE_SERIAL is best-effort, not all rpiboot devices have it set (some only show 32-bits)
 TARGET_DEVICE_SERIAL="$(udevadm info --name="$TARGET_DEVICE_PATH" --query=property --property=ID_SERIAL_SHORT --value)"
@@ -99,6 +100,19 @@ TARGET_DEVICE_SERIAL="$(udevadm info --name="$TARGET_DEVICE_PATH" --query=proper
 if [ -z "${TARGET_DEVICE_SERIAL}" ] || [ "${TARGET_DEVICE_SERIAL}" = "Broadcom" ]; then
     TARGET_DEVICE_SERIAL="${TARGET_DEVICE_PATH}"
     log "Using device path as serial: ${TARGET_DEVICE_SERIAL}"
+else
+    log "Using device serial: ${TARGET_DEVICE_SERIAL}"
+fi
+
+# If USB path extraction from udevadm failed, try using serial-based lookup as fallback
+if [ -z "${TARGET_USB_PATH}" ] && [ -n "${TARGET_DEVICE_SERIAL}" ] && [ "${TARGET_DEVICE_SERIAL}" != "${TARGET_DEVICE_PATH}" ]; then
+    log "USB path extraction from udevadm failed, attempting serial-based lookup"
+    TARGET_USB_PATH="$(get_usb_path_for_serial "${TARGET_DEVICE_SERIAL}" || true)"
+    if [ -n "${TARGET_USB_PATH}" ]; then
+        log "Found USB path via serial lookup: ${TARGET_USB_PATH}"
+    else
+        log "Warning: Could not determine USB path for device"
+    fi
 fi
 
 
@@ -232,23 +246,6 @@ derivePublicKey() {
     "${OPENSSL}" rsa -in "${CUSTOMER_KEY_FILE_PEM}" -pubout > "${CUSTOMER_PUBLIC_KEY_FILE}"
 }
 
-writeSig() {
-   SIG_TMP="$(mktemp)"
-   IMAGE="$1"
-   OUTPUT="$2"
-   sha256sum "${IMAGE}" | awk '{print $1}' > "${OUTPUT}"
-
-   # Include the update-timestamp
-   echo "ts: $(date -u +%s)" >> "${OUTPUT}"
-
-   if [ -n "$(get_signing_directives)" ]; then
-      # shellcheck disable=SC2046
-      "${OPENSSL}" dgst -sign $(get_signing_directives) -sha256 -out "${SIG_TMP}" "${IMAGE}"
-      echo "rsa2048: $(xxd -c 4096 -p < "${SIG_TMP}")" >> "${OUTPUT}"
-   fi
-   rm "${SIG_TMP}"
-}
-
 enforceSecureBootloaderConfig() {
     if ! grep -Fxq "SIGNED_BOOT=1" "${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}"; then
         echo "SIGNED_BOOT=1" >> "${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}"
@@ -271,6 +268,7 @@ identifyBootloaderConfig() {
 }
 
 # This function is adapted from the functions in the usbboot repo.
+# It handles both signed (secure-boot) and unsigned (naked) EEPROM updates
 update_eeprom() {
     src_image="$1"
     dst_image="$2"
@@ -292,7 +290,7 @@ update_eeprom() {
 
         TMP_CONFIG_SIG="$(mktemp)"
         log "Signing bootloader config"
-        writeSig "${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}" "${TMP_CONFIG_SIG}"
+        rpi-eeprom-digest -i "${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}" -o "${TMP_CONFIG_SIG}" -k "${pem_file}"
 
         # shellcheck disable=SC2086
         cat "${TMP_CONFIG_SIG}" ${DEBUG}
@@ -318,6 +316,9 @@ update_eeprom() {
                 cp "${src_image}" "${dst_image}.intermediate"
                 ;;
         esac
+    else
+        # Unsigned mode: just copy the source as intermediate
+        cp "${src_image}" "${dst_image}.intermediate"
     fi
 
     rm -f "${dst_image}"
@@ -349,11 +350,24 @@ BOOTLOADER_UPDATE_IMAGE=""
 BOOTLOADER_UPDATE_VERSION=0
 getBootloaderUpdateVersion() {
    BOOTLOADER_UPDATE_VERSION=0
+   
+   # Check if explicit firmware file is specified
+   if [ -n "${RPI_DEVICE_FIRMWARE_FILE}" ] && [ -f "${RPI_DEVICE_FIRMWARE_FILE}" ]; then
+      log "Using explicitly specified firmware file: ${RPI_DEVICE_FIRMWARE_FILE}"
+      BOOTLOADER_UPDATE_IMAGE="${RPI_DEVICE_FIRMWARE_FILE}"
+      BOOTLOADER_UPDATE_VERSION=$(strings "${RPI_DEVICE_FIRMWARE_FILE}" | grep BUILD_TIMESTAMP | sed 's/.*=//g')
+      return
+   fi
+   
+   # Fall back to automatic selection from release channel
    match=".*/pieeprom-[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9].bin"
    latest="$(find "${FIRMWARE_IMAGE_DIR}/" -maxdepth 1 -type f -follow -size "${EEPROM_SIZE}c" -regex "${match}" | sort -r | head -n1)"
    if [ -f "${latest}" ]; then
       BOOTLOADER_UPDATE_VERSION=$(strings "${latest}" | grep BUILD_TIMESTAMP | sed 's/.*=//g')
       BOOTLOADER_UPDATE_IMAGE="${latest}"
+   else
+      log "Warning: No firmware image found in ${FIRMWARE_IMAGE_DIR}/ matching size ${EEPROM_SIZE}c"
+      log "BOOTLOADER_UPDATE_IMAGE will remain unset - may cause issues if EEPROM update is required"
    fi
 }
 
@@ -398,22 +412,30 @@ case $TARGET_DEVICE_FAMILY in
 esac
 
 record_state "${TARGET_DEVICE_SERIAL}" "${BOOTSTRAP_STARTED}" "${TARGET_USB_PATH}"
+
+# Run bootstrap customisation script for the chosen provisioning style
+# This script runs when a device is detected, before provisioning begins
+case ${PROVISIONING_STYLE} in
+    "secure-boot")
+        run_customisation_script "sb-provisioner" "bootstrap" "${TARGET_DEVICE_SERIAL}" "${TARGET_DEVICE_FAMILY}" "${TARGET_USB_PATH}" "${TARGET_DEVICE_PATH}"
+        ;;
+    "fde-only")
+        run_customisation_script "fde-provisioner" "bootstrap" "${TARGET_DEVICE_SERIAL}" "${TARGET_DEVICE_FAMILY}" "${TARGET_USB_PATH}" "${TARGET_DEVICE_PATH}"
+        ;;
+    "naked")
+        run_customisation_script "naked-provisioner" "bootstrap" "${TARGET_DEVICE_SERIAL}" "${TARGET_DEVICE_FAMILY}" "${TARGET_USB_PATH}" "${TARGET_DEVICE_PATH}"
+        ;;
+    *)
+        log "Warning: Unknown provisioning style: ${PROVISIONING_STYLE}, skipping bootstrap customization"
+        ;;
+esac
+
 # Determine if we're enforcing secure boot, and if so, prepare the environment & eeprom accordingly.
 if [ "$ALLOW_SIGNED_BOOT" -eq 1 ]; then 
     if [ "${PROVISIONING_STYLE}" = "secure-boot" ]; then
         SECURE_BOOTLOADER_DIRECTORY="${RPI_SB_WORKDIR}/secure-bootloader/"
         mkdir -p "${SECURE_BOOTLOADER_DIRECTORY}"
         if [ -f "${SECURE_BOOTLOADER_DIRECTORY}/config.txt" ]; then
-            case ${TARGET_DEVICE_FAMILY} in
-                2712)
-                    BOOTCODE_BINARY_IMAGE="${SECURE_BOOTLOADER_DIRECTORY}/bootcode5.bin"
-                    BOOTCODE_FLASHING_NAME="${SECURE_BOOTLOADER_DIRECTORY}/bootcode5.bin"
-                    ;;
-                2711)
-                    BOOTCODE_BINARY_IMAGE="${SECURE_BOOTLOADER_DIRECTORY}/bootcode4.bin"
-                    BOOTCODE_FLASHING_NAME="${SECURE_BOOTLOADER_DIRECTORY}/bootcode4.bin"
-                    ;;
-            esac
             log "Secure bootloader directory already exists, skipping setup"
             if [ -f "/etc/rpi-sb-provisioner/special-reprovision-device/${TARGET_DEVICE_SERIAL}" ]; then
                 # This only makes sense if you're re-provisioning a device that's already been provisioned.
@@ -425,8 +447,30 @@ if [ "$ALLOW_SIGNED_BOOT" -eq 1 ]; then
                         die "No customer key file to use for re-provisioning. Aborting."
                     fi
                     log "Re-signing bootcode for special re-provisioning case"
+                    # For special reprovision, use the original firmware recovery.bin as source, not the cached signed bootcode
+                    BCM_CHIP=2712
+                    FIRMWARE_IMAGE_DIR="${FIRMWARE_ROOT}-${BCM_CHIP}/${FIRMWARE_RELEASE_STATUS}"
+                    BOOTCODE_BINARY_IMAGE="${FIRMWARE_IMAGE_DIR}/recovery.bin"
+                    BOOTCODE_FLASHING_NAME="${SECURE_BOOTLOADER_DIRECTORY}/bootcode5.bin"
                     rpi-sign-bootcode --debug -c 2712 -i "${BOOTCODE_BINARY_IMAGE}" -o "${BOOTCODE_FLASHING_NAME}" -k "${CUSTOMER_KEY_FILE_PEM}" -v 0 -n 16
+                else
+                    log "Warning: Special re-provisioning only supported on Pi 5 (2712), skipping for device family ${TARGET_DEVICE_FAMILY}"
                 fi
+            else
+                log "Normal provisioning mode (reusing cached bootloader)"
+            fi
+            if [ ! -f "/etc/rpi-sb-provisioner/special-reprovision-device/${TARGET_DEVICE_SERIAL}" ]; then
+                # Normal case: reuse cached bootcode
+                case ${TARGET_DEVICE_FAMILY} in
+                    2712)
+                        BOOTCODE_BINARY_IMAGE="${SECURE_BOOTLOADER_DIRECTORY}/bootcode5.bin"
+                        BOOTCODE_FLASHING_NAME="${SECURE_BOOTLOADER_DIRECTORY}/bootcode5.bin"
+                        ;;
+                    2711)
+                        BOOTCODE_BINARY_IMAGE="${SECURE_BOOTLOADER_DIRECTORY}/bootcode4.bin"
+                        BOOTCODE_FLASHING_NAME="${SECURE_BOOTLOADER_DIRECTORY}/bootcode4.bin"
+                        ;;
+                esac
             fi
             [ ! -f "/etc/rpi-sb-provisioner/special-skip-eeprom/${TARGET_DEVICE_SERIAL}" ] && timeout_fatal rpiboot -d "${SECURE_BOOTLOADER_DIRECTORY}" -p "${TARGET_USB_PATH}"
         else
@@ -435,7 +479,10 @@ if [ "$ALLOW_SIGNED_BOOT" -eq 1 ]; then
 
             announce_start "Setting up the environment for a signed-boot capable device"
             if [ -z "${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}" ]; then
-                RPI_DEVICE_BOOTLOADER_CONFIG_FILE=/var/lib/rpi-sb-provisioner/bootloader.default
+                RPI_DEVICE_BOOTLOADER_CONFIG_FILE=/var/lib/rpi-sb-provisioner/bootloader.secure
+                log "Using default secure bootloader config file: ${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}"
+            else
+                log "Using pre-configured bootloader config: ${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}"
             fi
 
             SOURCE_EEPROM_IMAGE=
@@ -485,8 +532,10 @@ if [ "$ALLOW_SIGNED_BOOT" -eq 1 ]; then
                         die "No Raspberry Pi EEPROM file to use as key vector"
                     else
                         update_eeprom "${SOURCE_EEPROM_IMAGE}" "${DESTINATION_EEPROM_IMAGE}" "${CUSTOMER_KEY_FILE_PEM}" "${CUSTOMER_PUBLIC_KEY_FILE}"
-                        writeSig "${DESTINATION_EEPROM_IMAGE}" "${DESTINATION_EEPROM_SIGNATURE}"
+                        rpi-eeprom-digest -i "${DESTINATION_EEPROM_IMAGE}" -o "${DESTINATION_EEPROM_SIGNATURE}" -k "${CUSTOMER_KEY_FILE_PEM}"
                     fi
+                else
+                    log "Using existing EEPROM signature: ${DESTINATION_EEPROM_SIGNATURE}"
                 fi
 
                 # This directive informs the bootloader to write the public key into OTP
@@ -510,7 +559,11 @@ if [ "$ALLOW_SIGNED_BOOT" -eq 1 ]; then
                         # Additionally, this only works on Raspberry Pi 5-family devices.
                         log "Re-signing bootcode for special re-provisioning case"
                         rpi-sign-bootcode --debug -c 2712 -i "${BOOTCODE_BINARY_IMAGE}" -o "${BOOTCODE_FLASHING_NAME}" -k "${CUSTOMER_KEY_FILE_PEM}" -v 0 -n 16
+                    else
+                        log "Warning: Special re-provisioning only supported on Pi 5 (2712), skipping for device family ${TARGET_DEVICE_FAMILY}"
                     fi
+                else
+                    log "Normal provisioning mode (not re-provisioning)"
                 fi
                 [ ! -f "/etc/rpi-sb-provisioner/special-skip-eeprom/${TARGET_DEVICE_SERIAL}" ] && timeout_fatal rpiboot -d "${SECURE_BOOTLOADER_DIRECTORY}" -p "${TARGET_USB_PATH}"
             else
@@ -528,101 +581,106 @@ if [ "$ALLOW_SIGNED_BOOT" -eq 1 ]; then
                     tar -vcf "${RPI_SB_WORKDIR}/bootfiles.bin" -- *
                     cd -
                     rm -rf "${FASTBOOT_SIGN_DIR}"
-
-                    announce_start "Signing fastboot image"
-                    cp "$(get_fastboot_gadget)" "${RPI_SB_WORKDIR}"/boot.img
-                    sha256sum "${RPI_SB_WORKDIR}"/boot.img | awk '{print $1}' > "${RPI_SB_WORKDIR}"/boot.sig
-                    printf 'rsa2048: ' >> "${RPI_SB_WORKDIR}"/boot.sig
-                    # Prefer PKCS11 over PEM keyfiles, if both are specified.
-                    # shellcheck disable=SC2046
-                    ${OPENSSL} dgst -sign $(get_signing_directives) -sha256 "${RPI_SB_WORKDIR}"/boot.img | xxd -c 4096 -p >> "${RPI_SB_WORKDIR}"/boot.sig
-                    cp "$(get_fastboot_config_file)" "${RPI_SB_WORKDIR}"/config.txt
-                    announce_stop "Signing fastboot image"
                     ;;
                 *)
                     # Raspberry Pi 4-class devices do not use signed bootcode files, so just copy the file into the relevant place.
                     cp /usr/share/rpiboot/mass-storage-gadget64/bootfiles.bin "${RPI_SB_WORKDIR}/bootfiles.bin"
                     ;;
             esac
+
+            announce_start "Signing fastboot image"
+            cp "$(get_fastboot_gadget)" "${RPI_SB_WORKDIR}"/boot.img
+            sha256sum "${RPI_SB_WORKDIR}"/boot.img | awk '{print $1}' > "${RPI_SB_WORKDIR}"/boot.sig
+            printf 'rsa2048: ' >> "${RPI_SB_WORKDIR}"/boot.sig
+            # Prefer PKCS11 over PEM keyfiles, if both are specified.
+            # shellcheck disable=SC2046
+            ${OPENSSL} dgst -sign $(get_signing_directives) -sha256 "${RPI_SB_WORKDIR}"/boot.img | xxd -c 4096 -p >> "${RPI_SB_WORKDIR}"/boot.sig
+            cp "$(get_fastboot_config_file)" "${RPI_SB_WORKDIR}"/config.txt
+            announce_stop "Signing fastboot image"
         fi
     else # !PROVISIONING_STYLE=secure-boot
-        NONSECURE_BOOTLOADER_DIRECTORY="${RPI_SB_WORKDIR}/non-secure-bootloader/"
-        mkdir -p "${NONSECURE_BOOTLOADER_DIRECTORY}"
-        if [ -f "${NONSECURE_BOOTLOADER_DIRECTORY}/config.txt" ]; then
-            log "Nonsecure bootloader directory already exists, skipping setup"
-        else
-            log "Creating nonsecure bootloader for future reuse"
-            touch "${NONSECURE_BOOTLOADER_DIRECTORY}/config.txt"
-            touch "${NONSECURE_BOOTLOADER_DIRECTORY}/bootcode.bin"
-            touch "${NONSECURE_BOOTLOADER_DIRECTORY}/pieeprom.bin"
-            touch "${NONSECURE_BOOTLOADER_DIRECTORY}/pieeprom.sig"
-            touch "${NONSECURE_BOOTLOADER_DIRECTORY}/bootcode4.bin"
-            touch "${NONSECURE_BOOTLOADER_DIRECTORY}/bootcode5.bin"
-
-            announce_start "Setting up the environment for a non-secure-boot capable device"
-            if [ -z "${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}" ]; then
-                RPI_DEVICE_BOOTLOADER_CONFIG_FILE=/var/lib/rpi-sb-provisioner/bootloader.default
-            fi
-
-            SOURCE_EEPROM_IMAGE=
-            DESTINATION_EEPROM_IMAGE=
-            DESTINATION_EEPROM_SIGNATURE=
-            BOOTCODE_BINARY_IMAGE=
-            BOOTCODE_FLASHING_NAME=
-            case $TARGET_DEVICE_FAMILY in
-                2711)
-                    BCM_CHIP=2711
-                    EEPROM_SIZE=524288
-                    FIRMWARE_IMAGE_DIR="${FIRMWARE_ROOT}-${BCM_CHIP}/${FIRMWARE_RELEASE_STATUS}"
-                    getBootloaderUpdateVersion
-                    SOURCE_EEPROM_IMAGE="${BOOTLOADER_UPDATE_IMAGE}"
-                    BOOTCODE_BINARY_IMAGE="${FIRMWARE_IMAGE_DIR}/recovery.bin"
-                    BOOTCODE_FLASHING_NAME="${NONSECURE_BOOTLOADER_DIRECTORY}/bootcode4.bin"
-                    ;;
-                2712)
-                    BCM_CHIP=2712
-                    EEPROM_SIZE=2097152
-                    FIRMWARE_IMAGE_DIR="${FIRMWARE_ROOT}-${BCM_CHIP}/${FIRMWARE_RELEASE_STATUS}"
-                    getBootloaderUpdateVersion
-                    SOURCE_EEPROM_IMAGE="${BOOTLOADER_UPDATE_IMAGE}"
-                    BOOTCODE_BINARY_IMAGE="${FIRMWARE_IMAGE_DIR}/recovery.bin"
-                    BOOTCODE_FLASHING_NAME="${NONSECURE_BOOTLOADER_DIRECTORY}/bootcode5.bin"
-                    ;;
-                *)
-                    record_state "${TARGET_DEVICE_SERIAL}" "${BOOTSTRAP_ABORTED}" "${TARGET_USB_PATH}"
-                    die "Unable to identify EEPROM parameters for non-Pi4, Pi5 device. Aborting."
-            esac
-
-            DESTINATION_EEPROM_IMAGE="${NONSECURE_BOOTLOADER_DIRECTORY}/pieeprom.bin"
-
-            ### In the completely-unprovisioned state, where you have not yet written a customer OTP key, simply make the copy of the unsigned bootcode
-            cp "${BOOTCODE_BINARY_IMAGE}" "${BOOTCODE_FLASHING_NAME}"
-            ####
-
-            # This directive tells the bootloader to reboot once it's written the OTP
-            echo "recovery_reboot=1" >> "${NONSECURE_BOOTLOADER_DIRECTORY}/config.txt"
-
-            if [ -n "${RPI_DEVICE_LOCK_JTAG}" ]; then
-                log "JTAG lock requested, but not supported on non-secure-boot devices"
-            fi
-
-            if [ -n "${RPI_DEVICE_EEPROM_WP_SET}" ]; then
-                log "EEPROM write-protect requested, but not supported on non-secure-boot devices"
-            fi
-
-            log "Writing key and EEPROM configuration to the device"
-            # shellcheck disable=SC2086
-            if ! rpi-eeprom-config \
-                --config "${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}" \
-                --out "${DESTINATION_EEPROM_IMAGE}" \
-                "${SOURCE_EEPROM_IMAGE}"; then
-                record_state "${TARGET_DEVICE_SERIAL}" "${BOOTSTRAP_ABORTED}" "${TARGET_USB_PATH}"
-                die "Failed to update EEPROM image"
-            fi
-            [ ! -f "/etc/rpi-sb-provisioner/special-skip-eeprom/${TARGET_DEVICE_SERIAL}" ] && timeout_fatal rpiboot -d "${NONSECURE_BOOTLOADER_DIRECTORY}" -p "${TARGET_USB_PATH}"
-        fi
+        # For non-secure-boot provisioning on 2711/2712, still update EEPROM but without signing
         case ${TARGET_DEVICE_FAMILY} in
             2712|2711)
+                NON_SECURE_BOOTLOADER_DIRECTORY="${RPI_SB_WORKDIR}/non-secure-bootloader/"
+                mkdir -p "${NON_SECURE_BOOTLOADER_DIRECTORY}"
+                
+                if [ ! -f "${NON_SECURE_BOOTLOADER_DIRECTORY}/config.txt" ]; then
+                    log "Creating non-secure bootloader for EEPROM update"
+                    touch "${NON_SECURE_BOOTLOADER_DIRECTORY}/config.txt"
+                    
+                    announce_start "Setting up EEPROM update for non-secure-boot device"
+                    if [ -z "${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}" ]; then
+                        RPI_DEVICE_BOOTLOADER_CONFIG_FILE=/var/lib/rpi-sb-provisioner/bootloader.naked
+                        log "Using default naked bootloader config file: ${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}"
+                    else
+                        log "Using pre-configured bootloader config: ${RPI_DEVICE_BOOTLOADER_CONFIG_FILE}"
+                    fi
+                    
+                    SOURCE_EEPROM_IMAGE=
+                    DESTINATION_EEPROM_IMAGE=
+                    DESTINATION_EEPROM_SIGNATURE=
+                    BOOTCODE_BINARY_IMAGE=
+                    BOOTCODE_FLASHING_NAME=
+                    
+                    case $TARGET_DEVICE_FAMILY in
+                        2711)
+                            BCM_CHIP=2711
+                            EEPROM_SIZE=524288
+                            FIRMWARE_IMAGE_DIR="${FIRMWARE_ROOT}-${BCM_CHIP}/${FIRMWARE_RELEASE_STATUS}"
+                            getBootloaderUpdateVersion
+                            SOURCE_EEPROM_IMAGE="${BOOTLOADER_UPDATE_IMAGE}"
+                            BOOTCODE_BINARY_IMAGE="${FIRMWARE_IMAGE_DIR}/recovery.bin"
+                            BOOTCODE_FLASHING_NAME="${NON_SECURE_BOOTLOADER_DIRECTORY}/bootcode4.bin"
+                            ;;
+                        2712)
+                            BCM_CHIP=2712
+                            EEPROM_SIZE=2097152
+                            FIRMWARE_IMAGE_DIR="${FIRMWARE_ROOT}-${BCM_CHIP}/${FIRMWARE_RELEASE_STATUS}"
+                            getBootloaderUpdateVersion
+                            SOURCE_EEPROM_IMAGE="${BOOTLOADER_UPDATE_IMAGE}"
+                            BOOTCODE_BINARY_IMAGE="${FIRMWARE_IMAGE_DIR}/recovery.bin"
+                            BOOTCODE_FLASHING_NAME="${NON_SECURE_BOOTLOADER_DIRECTORY}/bootcode5.bin"
+                            ;;
+                    esac
+                    
+                    DESTINATION_EEPROM_IMAGE="${NON_SECURE_BOOTLOADER_DIRECTORY}/pieeprom.bin"
+                    DESTINATION_EEPROM_SIGNATURE="${NON_SECURE_BOOTLOADER_DIRECTORY}/pieeprom.sig"
+                    
+                    # Copy unsigned bootcode
+                    cp "${BOOTCODE_BINARY_IMAGE}" "${BOOTCODE_FLASHING_NAME}"
+                    
+                    # Update EEPROM using the standard update_eeprom function
+                    # Calling without pem_file creates an unsigned EEPROM update
+                    if [ ! -e "${DESTINATION_EEPROM_SIGNATURE}" ]; then
+                        if [ ! -e "${SOURCE_EEPROM_IMAGE}" ]; then
+                            record_state "${TARGET_DEVICE_SERIAL}" "${BOOTSTRAP_ABORTED}" "${TARGET_USB_PATH}"
+                            die "No Raspberry Pi EEPROM file found for update"
+                        else
+                            log "Updating EEPROM with bootloader configuration (unsigned)"
+                            # Call update_eeprom without signing parameters
+                            update_eeprom "${SOURCE_EEPROM_IMAGE}" "${DESTINATION_EEPROM_IMAGE}" "" ""
+                            
+                            # Create unsigned signature using rpi-eeprom-digest
+                            # Without -k parameter, creates hash + timestamp only
+                            log "Creating unsigned pieeprom.sig using rpi-eeprom-digest"
+                            rpi-eeprom-digest -i "${DESTINATION_EEPROM_IMAGE}" -o "${DESTINATION_EEPROM_SIGNATURE}"
+                        fi
+                    else
+                        log "Using existing EEPROM signature: ${DESTINATION_EEPROM_SIGNATURE}"
+                    fi
+                    
+                    # Simple recovery config to update EEPROM and reboot
+                    echo "recovery_reboot=1" > "${NON_SECURE_BOOTLOADER_DIRECTORY}/config.txt"
+                    
+                    log "Updating EEPROM to latest version"
+                    [ ! -f "/etc/rpi-sb-provisioner/special-skip-eeprom/${TARGET_DEVICE_SERIAL}" ] && timeout_fatal rpiboot -d "${NON_SECURE_BOOTLOADER_DIRECTORY}" -p "${TARGET_USB_PATH}"
+                    log "EEPROM update completed. Device rebooted."
+                else
+                    log "Reusing existing non-secure bootloader configuration"
+                fi
+                
+                # Prepare fastboot files
                 cp /usr/share/rpiboot/mass-storage-gadget64/bootfiles.bin "${RPI_SB_WORKDIR}/bootfiles.bin"
                 cp "$(get_fastboot_gadget)" "${RPI_SB_WORKDIR}"/boot.img
                 cp "$(get_fastboot_config_file)" "${RPI_SB_WORKDIR}"/config.txt
@@ -653,6 +711,9 @@ if [ -n "${TARGET_DEVICE_SERIAL}" ]; then
         mv "${early_log_dir}/metadata" "${target_log_dir}/metadata"
     fi
     mv "${early_log_dir}/bootstrap.log" "${target_log_dir}/bootstrap.log"
+else
+    log "Warning: TARGET_DEVICE_SERIAL is empty, logs will remain in ${EARLY_LOG_DIRECTORY}"
+    echo "[$(date +"%Y-%m-%d %H:%M:%S.$(date +%N | cut -c1-3)")] Warning: TARGET_DEVICE_SERIAL is empty, logs will remain in ${EARLY_LOG_DIRECTORY}" >> "${EARLY_LOG_DIRECTORY}/bootstrap.log"
 fi
 
 announce_stop "fastboot initialisation"
