@@ -48,7 +48,7 @@ die() {
 }
 
 simg_expanded_size() {
-    echo "$(($(simg_dump "$1" | sed -E 's/.*?Total of ([0-9]+) ([0-9]+)-byte .*/\1 * \2/')))"
+    echo "$(($(simg_dump "$1" | sed -E 's/.*Total of ([0-9]+) ([0-9]+)-byte .*/\1 * \2/')))"
 }
 
 log() {
@@ -441,7 +441,7 @@ augment_initramfs() {
     cd -
 
     # Generate depmod information
-    for kernel in $(find "${initramfs_dir}usr/lib/modules" -mindepth 1 -maxdepth 1 -type d -printf '%f\n'); do
+    find "${initramfs_dir}usr/lib/modules" -mindepth 1 -maxdepth 1 -type d -exec basename {} \; | while read -r kernel; do
         depmod --basedir "${initramfs_dir}" "${kernel}"
     done
 
@@ -626,15 +626,15 @@ fastboot -s "${FASTBOOT_DEVICE_SPECIFIER}" flash "${RPI_DEVICE_STORAGE_TYPE}"p1
 fastboot -s "${FASTBOOT_DEVICE_SPECIFIER}" flash mapper/cryptroot "${RPI_SB_WORKDIR}"/rootfs-temporary.simg
 announce_stop "Writing OS images"
 
+metadata_gather
+
 # Run customisation script for post-flash stage
-run_customisation_script "sb-provisioner" "post-flash"
+run_customisation_script "sb-provisioner" "post-flash" "${FASTBOOT_DEVICE_SPECIFIER}" "${TARGET_DEVICE_SERIAL}" "${RPI_DEVICE_STORAGE_TYPE}"
 
 announce_start "Set LED status"
 fastboot -s "${FASTBOOT_DEVICE_SPECIFIER}" oem led PWR 0
 announce_stop "Set LED status"
 
-metadata_gather
-
 record_state "${TARGET_DEVICE_SERIAL}" "${PROVISIONER_FINISHED}" "${TARGET_USB_PATH}"
 log "Provisioning completed. Remove the device from this machine."
 
